/*
 * data.c
 *
 * application data
 *
 * 23/08/2017    GF    Created
 * 
 * 
 * MIT License
 * 
 * Copyright 2017 Graeme Foot (Kinetic Engineering Design Ltd.)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */


#include <stdint.h>
#include <stdbool.h>

#include <rtai_types.h>
#include <rtai_lxrt.h>


#include "data.h"


static int64_t u_appTimeBase = 0LL;


/** get the rt time in ns for the current cpu (adjusted for the app time base)
 * 
 * \ret the time in ns
 */
uint64_t app_getTimeNS()
{
  RTIME time = rt_get_time_ns();
  
  if (u_appTimeBase > time)
  {
    //logErr(VERBOSITY_CRITICAL, "app_getTimeNS error: TimeBase greater than system time (timeBase: %lld, sysTime: %llu\n",
    //       u_appTimeBase, time);
    return time;
  }
  else
  {
    return time - u_appTimeBase;
  }
}



/** convert app time to rtai time in counts (via the app time base)
 */
RTIME app2rtaiTime(
        uint64_t in_appTime
        )
{
  RTIME time;
  
  if ((u_appTimeBase < 0) && ((uint64_t)(-u_appTimeBase) > in_appTime))
  {
    //logErr(VERBOSITY_CRITICAL, "APP: app2rtaiTime error: TimeBase less than app time (timeBase: %lld, appTime: %llu\n",
    //       u_appTimeBase, in_appTime);
    time = in_appTime;
  }
  else
  {
    time = in_appTime + u_appTimeBase;
  }
  
  return nano2count(time);
}




/** add to the app time base value
 * 
 * if the app clock is slower (ie the period takes longer) then the time base
 * value should be increased each period
 */
void app_addTimeBase(
        int64_t in_timeBase
        )
{
  u_appTimeBase += in_timeBase;
}



/**
 * log message to syslog
 */
void logMsg(const char *__format, ...)
{
  // copy of what lxrt does with rt_printk
	char display[VSNPRINTF_BUF_SIZE];
	struct { const char *display; long nch; } arg = { display, 0 };
  
  // prefix
  arg.nch = snprintf(display, VSNPRINTF_BUF_SIZE-2, "APP: ");
  
  // add args
	va_list args;
	va_start(args, __format);

  arg.nch += vsnprintf(&display[arg.nch], VSNPRINTF_BUF_SIZE-2, __format, args);
	
  va_end(args);
	
  
  // force eol
  if (display[arg.nch-1] != '\n') // 10u
  {
    display[arg.nch]   = '\n';    // 10u;
    display[arg.nch+1] = '\0';    // 0u;

    arg.nch++;
  }
  

  rtai_lxrt(BIDX, SIZARG, PRINTK, &arg);
  
  
	return;
}




/** wait for the next period
 */
void ecat_waitPeriod()
{
  while (true)
  {
    // get wakeTime in timer counts
    // Note: not in nano-seconds!!!!
    RTIME wakeTime = app2rtaiTime(g_app.nextWakeTime);
    
    switch (rt_sleep_until(wakeTime))
    {
      case RTE_UNBLKD : 
      {
        logMsg("sysMain_waitPeriod = RTE_UNBLKD\n");
        continue;
      }
      
      case RTE_TMROVRN :
      {
        logMsg("sysMain_waitPeriod = RTE_TMROVRN\n");
        g_app.overruns++;
        
        if (g_app.overruns % 100 == 0)
        {
          // cycle time might be stuffed, so ensure other processes get some
          // time slice
          rt_sleep(10000);
        }
      } break;
      
      default : {}
    }
    
    // done if we got to here
    break;
  }


  // statistics and header data
  g_app.actualWakeTime  = app_getTimeNS();
  
  
  // calc next wake time (in app time)
  g_app.nextWakeTime += g_app.scanTimeNS;
}







