/*
 * etherCATMaster.c
 *
 * implementation of an EtherCAT master
 *
 * 23/08/2017    GF    Created
 * 
 * 
 * MIT License
 * 
 * Copyright 2017 Graeme Foot (Kinetic Engineering Design Ltd.)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <assert.h>
#include <ecrt.h>
#include <asm/rtai_hal.h>


#include "data.h"
#include "etherCATMaster.h"


/** return the sign of a number
 *
 * ie -1 for -ve value, 0 for 0, +1 for +ve value
 *
 * \retval the sign of the value
 */
#define sign(val) \
({ typeof (val) _val = (val); \
   ((_val > 0) - (_val < 0)); })



/** return a number between the two given numbers
 *
 * Note: a define so that it can accept all data types
 *
 * \retval the value in the range
 */
#define ensureRange(value, minVal, maxVal) \
({ typeof (value)  _value  = (value); \
   typeof (minVal) _minVal = (minVal); \
   typeof (maxVal) _maxVal = (maxVal); \
   (_value < _minVal ? _minVal : (_value > _maxVal ? _maxVal : _value)); })




/** Calc the EtherCAT item data address
 */
void setECAddress(
        ecItem_s *ecItem
        )
{
  ecItem->addr = ecItem->master->domains[ecItem->domain].data + ecItem->offset;
}



/** synchronise the distributed clock
 */
static int ecMaster_syncDistClock(
        void *this            /**< pointer to module module_s */
        )
{
  assert(this);
  
  etherCATMaster_s *ecMaster = this;
  
  
  // get lower 32 bit of clock time from reference slave (after first scan)
  if (ecMaster->m_getDCDiff)
  {
    int      res;
    uint32_t slaveTime;
    res = ecrt_master_reference_clock_time(ecMaster->master, &slaveTime);

    switch (res)
    {
      case 0 :
      {
        // calc time diff with prev master time
        ecMaster->m_dcDiff = (uint32_t)ecMaster->m_dcTime - slaveTime;

        // let user know error returned to normal
        if (ecMaster->m_lastDCRes != res)
        { 
           logMsg("Reference clock time resumed");
        }
      } break;

      default :
      {
        // report first error
        if (ecMaster->m_lastDCRes != res)
        {
           logMsg("Failed to get Reference clock time: %s", strerror(-(res)));
        }

        // no ref clock found or datagram failure
        ecMaster->m_dcDiff = 0;
      }
    }
    ecMaster->m_lastDCRes = res;
  }
  else
  {
    ecMaster->m_dcDiff    = 0;
    ecMaster->m_getDCDiff = true;
  }

  // call to sync slaves to ref slave 
  // (which is used for ecrt_master_reference_clock_time)
  ecrt_master_sync_slave_clocks(ecMaster->master);

  // queue the 64bit ref slave clock time datagram
  // Note: we won't actually read the value, we want it for wireshark analysis
  ecrt_master_64bit_reference_clock_time_queue(ecMaster->master);

  // set master time in nano-seconds
  ecMaster->m_dcTime = app_getTimeNS();
  ecrt_master_application_time(ecMaster->master, ecMaster->m_dcTime);
  
  
  return 0;
}



// dc clock master variables
// the first ecMaster will automatically become the dc primary

#define DC_FILTER_CNT 1024

static bool     u_dcStarted = false;
static int32_t  u_dcLastDiff = 0;
static int64_t  u_dcDiffTot = 0LL;
static int64_t  u_dcDeltaTot = 0LL;
static int      u_dcIdx = 0;


/** update the master time based on ref slaves time diff
 * 
 * called after the ethercat frame is sent to avoid time jitter in
 * ecMaster_syncDistClock()
 */
static int ecMaster_updateMasterClock(
        void *this            /**< pointer to module module_s */
        )
{
  assert(this);
  
  etherCATMaster_s *ecMaster = this;
  
  assert(ecMaster->master);
  
  
  if (u_dcStarted)
  {
    // calc drift (via un-normalised time diff)
    g_app.rtTimeDelta = ecMaster->m_dcDiff - u_dcLastDiff;
    u_dcLastDiff  = ecMaster->m_dcDiff;

    // normalise the time diff
    // the ethercat code can be one cycle out, but we just want to stay synced
    // to the nearest cycle
    ecMaster->m_dcDiff = ((ecMaster->m_dcDiff + (g_app.scanTimeNS/2)) % g_app.scanTimeNS) - (g_app.scanTimeNS/2);
    g_app.rtTimeDiff    = ecMaster->m_dcDiff;
    
    // update current time
    g_app.rtTimeCurr = ecMaster->m_dcTime;

    
    // add to totals
    u_dcDiffTot  += g_app.rtTimeDiff;
    u_dcDeltaTot += g_app.rtTimeDelta;
    u_dcIdx++;
    
    if (u_dcIdx >= DC_FILTER_CNT)
    {
      // add rounded delta average
      g_app.rtAdjBy += ((u_dcDeltaTot + (DC_FILTER_CNT/2)) / DC_FILTER_CNT);
      
      // and add adjustment for general diff
      g_app.rtAdjBy += sign(u_dcDiffTot / DC_FILTER_CNT);
      
      // limit crazy numbers (0.1% of std cycle time)
      g_app.rtAdjBy = ensureRange(g_app.rtAdjBy, -1000, 1000);
      
      // reset
      u_dcDiffTot  = 0LL;
      u_dcDeltaTot = 0LL;
      u_dcIdx = 0;
    }
    
    
    // add cycles adjustment to time base (including a spot adjustment)
    app_addTimeBase(g_app.rtAdjBy + sign(ecMaster->m_dcDiff));
  }
  else
  {
    u_dcStarted = (ecMaster->m_dcDiff != 0);
    
    if (u_dcStarted)
    {
      // output first diff
      logMsg("first master diff: %d.\n", ecMaster->m_dcDiff);
      
      
      // cache first diff for delta calcs
      g_app.rtTimeDelta = 0LL;
      u_dcLastDiff  = ecMaster->m_dcDiff;
      
      // normalise the time diff
      ecMaster->m_dcDiff = ((ecMaster->m_dcDiff + (g_app.scanTimeNS/2)) % g_app.scanTimeNS) - (g_app.scanTimeNS/2);
      g_app.rtTimeDiff    = ecMaster->m_dcDiff;

      // record the time of this initial cycle
      g_app.rtTimeStart = ecMaster->m_dcTime;
    }
  }
  
  
  return 0;
}



/** allocate module IO function pointer definition
 *
 * \retval 0 on success, else error code
 */
static int ecMaster_prepareToRun( 
        void *this                /**< pointer to module module_s */
        )  
{
  assert(this);
  
  etherCATMaster_s *ecMaster = this;

  
  ecDevice_s        *dev;
  int                ecRes;
  int                i;
  ec_slave_config_t *refSlaveConfig = NULL;
  
  
  // pre set up domain memory in non-rt
  ecRes = ecrt_master_setup_domain_memory(ecMaster->master);
  if (ecRes < 0)
  {
    logMsg("EtherCAT - ACTIVATE_ERROR, setup domain memory error %d", ecRes);
    return -1;
  }

  
  // prepare domain memory
  for (i = 0; i < EC_DOMAIN_COUNT; i++)
  {
    // get domain PDO data addresses and the domain size
    ecMaster->domains[i].data = ecrt_domain_data(ecMaster->domains[i].domain);
    ecMaster->domains[i].size = ecrt_domain_size(ecMaster->domains[i].domain);
    if ((ecMaster->domains[i].data == NULL) && (ecMaster->domains[i].size > 0))
    {
      logMsg("Domain data error");
      return -1;
    }
  }

  
  // iterate slave devices and call prepare to run
  for (i = 0; i < ecMaster->deviceCount; i++)
  {
    dev = ecMaster->devices[i];
    if (dev->prepareToRun(dev) != 0) return -1;
  }
  
  
  // set up distributed clocks
  // get initial application time, in phase with realtime cycle
  ecMaster->m_dcTimeStart  = app_getTimeNS();
  ecMaster->m_dcTimeStart -= (ecMaster->m_dcTimeStart % g_app.scanTimeNS);
  ecMaster->m_dcTime       = ecMaster->m_dcTimeStart;
  

  if (ecMaster->m_dcRefSlave)
  {
    refSlaveConfig = ecMaster->m_dcRefSlave->slaveConfig;
  }
  
  // set initial app time
  ecrt_master_application_time(ecMaster->master, ecMaster->m_dcTimeStart);

  // select the slave device to be the reference clock
  ecRes = ecrt_master_select_reference_clock(ecMaster->master, refSlaveConfig);
  if (ecRes < 0)
  {
    logMsg("EtherCAT - ACTIVATE_ERROR, setup distributed clock error %d", ecRes);
    return -1;
  }

  // activate the master
  ecRes = ecrt_master_activate(ecMaster->master);
  if (ecRes < 0)
  {
    logMsg("EtherCAT - ACTIVATE_ERROR, master activate error %d", ecRes);
    return -1;
  }
  
  // set active flag
  ecMaster->masterActivateState = maActive;
  
  
  return 0;
}



/** check the domain state is happy
 */
static int ecMaster_checkDomainState(
        void *this              /**< pointer to module module_s */
        )
{
  assert(this);
  assert(isClass(this, CEtherCATModule));
  
  etherCATMaster_s *ecMaster = this;

  assert(ecMaster->master);
  
  ec_domain_state_t  newState;
  int                i;
  

  for (i = 0; i < EC_DOMAIN_COUNT; i++)
  {
    // get current state
    ecrt_domain_state(ecMaster->domains[i].domain, &newState);

    // log changes
    if (newState.working_counter != ecMaster->domains[i].state.working_counter)
    {
      logMsg("EtherCAT Master: Domain %d: WC %u.\n", i, newState.working_counter);
    }
    if (newState.wc_state != ecMaster->domains[i].state.wc_state)
    {
      logMsg("EtherCAT Master: Domain %d: State %u.\n", i, newState.wc_state);
    }

    // cache state
    ecMaster->domains[i].state = newState;
  }
  
  
  return 0;
}



/** check the master state is happy
 */
static int ecMaster_checkMasterState(
        void *this            /**< pointer to module module_s */
        )
{
  assert(this);
  
  etherCATMaster_s *ecMaster = this;

  assert(ecMaster->master >= 0);
  
  ec_master_state_t  newState;
  

  ecrt_master_state(ecMaster->master, &newState);

  
  if (newState.slaves_responding != ecMaster->masterState.slaves_responding)
  {
     logMsg("EtherCAT Master: %u slave(s).\n", newState.slaves_responding);
  }
  if (newState.al_states != ecMaster->masterState.al_states)
  {
    logMsg("EtherCAT Master: AL states: 0x%02X.\n", newState.al_states);
  }
  if (newState.link_up != ecMaster->masterState.link_up)
  {
    logMsg("EtherCAT Master: Link is %s.\n", newState.link_up ? "up" : "down");
  }

  ecMaster->masterState = newState;
  
  
  return 0;
}



/** write the module
 */
static int ecMaster_run(
        void *this               /**< pointer to the module module_s */
        )
{
  etherCATMaster_s *ecMaster = this;

  
  // receive process data
  ecrt_master_receive(ecMaster->master);

  // process domain data
  for (int i = 0; i < EC_DOMAIN_COUNT; i++)
  {
    ecrt_domain_process(ecMaster->domains[i].domain);
  }


  // check domain happiness factor
  if (ecMaster_checkDomainState(ecMaster) != 0) return -1;

  // check master happiness factor
  if (ecMaster->checkMasterCntr <= 0)
  {
    ecMaster->checkMasterCntr = 10;
    if (ecMaster_checkMasterState(ecMaster) != 0) return -1;
  }
  ecMaster->checkMasterCntr--;
  


  
  // run device logic if we are active
  if (ecMaster->masterActivateState == maActive)
  {
    // get the devices to run any internal logic 
    ecDevice_s *ecDev;
    for (int i = 0; i < ecMaster->deviceCount; i++)
    {
      ecDev = ecMaster->devices[i];
      ecDev->run(ecDev);
    }
  }

  
  
  
  // send process data
  for (int i = 0; i < EC_DOMAIN_COUNT; i++)
  {
    ecrt_domain_queue(ecMaster->domains[i].domain);
  }

  // we always sync distributed clock (just before master_send)
  if (ecMaster_syncDistClock(ecMaster) != 0) return -1;
  
  // do the send
  ecrt_master_send(ecMaster->master);   
  
  
  // update the master clock (if this module provides the ref slave)
  // Note: called after ecrt_master_send to reduce time jitter
  if (ecMaster_updateMasterClock(ecMaster) != 0) return -1;



  // thats all folks
  return 0;
}



/** prepare the module to stop
 *
 * \retval 0 on success else error code
 */
static int ecMaster_prepareToStop(
        void *this                /**< pointer to module module_s */
        )
{
  assert(this);

  etherCATMaster_s *ecMaster = this;
  ecDevice_s       *ecDev;
  
  int i;
  
  // stop the devices 
  for (i = 0; i < ecMaster->deviceCount; i++)
  {
    ecDev = ecMaster->devices[i];
    ecDev->prepareToStop(ecDev);
  }
  
  // deactivate slaves, to clear dc clock
  ecrt_master_deactivate_slaves(ecMaster->master);

  
  return 0;
}



/** prepare the module to stop
 *
 * \retval 0 on success else error code
 */
static int ecMaster_stop(
        void *this                /**< pointer to module module_s */
        )
{
  assert(this);

  etherCATMaster_s *ecMaster = this;


  // deactivate the master
  ecrt_master_deactivate(ecMaster->master);

  // release the master
  ecrt_release_master(ecMaster->master);
  ecMaster->master = NULL;
  

  return 0;
}

  
  
/** create / initialise an EtherCAT module
 *
 * \retval 0 on success else error code
 */
int create_etherCATMaster(
        ecDeviceList_t *in_devices,     /**< list of devices to configure */
        void **out_module               /**< a pointer to the new module */
        )
{
  // count devices
  ecDeviceList_t *devItem;
  int             devCnt = 0;
  for (devItem = &in_devices[0]; (devItem->deviceTypeName != NULL); devItem++)
  {
    devCnt++;
  }
  
  
  // alloc memory, devices list at the end
  void *this = calloc(1, sizeof(etherCATMaster_s) + (sizeof(void) * devCnt));
  etherCATMaster_s *ecMaster = this;
  
  
  // set methods
  ecMaster->prepareToRun  = ecMaster_prepareToRun;
  ecMaster->run           = ecMaster_run;
  ecMaster->prepareToStop = ecMaster_prepareToStop;
  ecMaster->stop          = ecMaster_stop;
  
  
  
  // request the master
  ecMaster->master = ecrt_request_master(0);
  if (ecMaster->master == NULL)
  {
    logMsg("Master not available");
    return -1;
  }


  // create the domains
  int i;
  for (i = 0; i < EC_DOMAIN_COUNT; i++)
  {
    ecMaster->domains[i].size   = 0;
    ecMaster->domains[i].data   = NULL;
    ecMaster->domains[i].domain = ecrt_master_create_domain(ecMaster->master);
    if (ecMaster->domains[i].domain == NULL)
    {
      logMsg("Error creating domain");
      return -1;
    }
  }
  

  // create devices
  ecMaster->devices     = this + sizeof(etherCATMaster_s);
  ecMaster->deviceCount = 0;
  
  ecDeviceInfo_t *devInfoItem;
  for (devItem = &in_devices[0]; (devItem->deviceTypeName != NULL); devItem++)
  {
    // lookup dev info
    for (devInfoItem = &ecDeviceTypes[0];  (devInfoItem->deviceTypeName != NULL);  devInfoItem++)
    {
      if (strcmp(devItem->deviceTypeName, devInfoItem->deviceTypeName) == 0)
      {
        break;
      }
    }
    
    // create the device
    if (devInfoItem->deviceTypeName != NULL)
    {
      ecMaster->devices[ecMaster->deviceCount] = calloc(1, devInfoItem->dataSize);
      
      ecMaster->devices[ecMaster->deviceCount]->m_parent    = ecMaster;
      ecMaster->devices[ecMaster->deviceCount]->alias       = devItem->alias;
      ecMaster->devices[ecMaster->deviceCount]->position    = devItem->position;
      ecMaster->devices[ecMaster->deviceCount]->vendorID    = devInfoItem->vendorID;
      ecMaster->devices[ecMaster->deviceCount]->productCode = devInfoItem->productCode;
      ecMaster->devices[ecMaster->deviceCount]->subCode     = devInfoItem->subCode;
      ecMaster->devices[ecMaster->deviceCount]->revision    = 0;
      
      devInfoItem->create(ecMaster->devices[ecMaster->deviceCount]);
      
      // is this the dc ref slave?
      if (devItem->dcRefSlave)
      {
        ecMaster->m_dcRefSlave = ecMaster->devices[ecMaster->deviceCount];
      }
      
      ecMaster->deviceCount++;
    }
  }
  

  
  // init values
  ecMaster->m_lastDCRes = 0;
  
  
  *out_module = this;
  

  return 0;
}




