/*
 * etherCATMaster.h
 *
 * implementation of an EtherCAT master
 *
 * 23/08/2017    GF    Created
 * 
 * 
 * MIT License
 * 
 * Copyright 2017 Graeme Foot (Kinetic Engineering Design Ltd.)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef _ETHERCAT_MASTER_H
#define	_ETHERCAT_MASTER_H


#include <stdbool.h>
#include <ecrt.h>
#include <rtdm/rtdm.h>



// forward declarations
typedef struct _etherCATMaster_s etherCATMaster_s;
typedef struct _ecDevice_s ecDevice_s;



/** The EtherCAT domain data structure
 */
typedef struct _etherCATDomain_t
{
  ec_domain_t        *domain;       /**< the domain pointer */
  size_t              size;         /**< the size of the primary domain */
  void               *data;         /**< the primary domain PDO data address */
  ec_domain_state_t   state;        /**< cached domain state */
} etherCATDomain_t;



/** EtherCAT data ref
 */
typedef struct _ecItem_s
{
  char             *name;
  etherCATMaster_s *master;
  int               domain;
  int               offset;
  unsigned int      bitPos;
  void             *addr;
} ecItem_s;


/** device function
 *
 * \retval 0 on success, else error code
 */
typedef int (*ecDevice_create_fn)(
        void *this               /**< the device struct */
        );


/** info for known device types
 */
typedef struct _ecDeviceInfo_t
{
  char         *deviceTypeName;       /**< device type name from config file */
  uint32_t      vendorID;             /**< device vendor ID */
  uint32_t      productCode;          /**< device product code */
  uint32_t      subCode;              /**< module sub code type */
  size_t        dataSize;             /**< struct size, must be at least sizeof(ecDevice_s) */
  
  ecDevice_create_fn create;          /**< specific device creation */
  
} ecDeviceInfo_t;


/** database of known device types
 */
extern ecDeviceInfo_t ecDeviceTypes[];



/** device function pointer
 *
 * \retval 0 on success else -1
 */
typedef int (*ecDevice_fn)(
        void *this                      /**< the device ecDevice_s */
        );



/** Base EtherCAT slave device data structure
 */
struct _ecDevice_s
{
  etherCATMaster_s   *m_parent;         /**< parent EtherCAT master */

  // EtherCAT data
  uint16_t            alias;            /**< ethercat device alias 
                                         *   (0 means position = absolute position) */
  uint16_t            position;         /**< ethercat device position 
                                         *   (absolute pos if alias = 0, or relative pos if alias > 0 */
  
  uint32_t            vendorID;         /**< device vendor ID */
  uint32_t            productCode;      /**< device product code */
  uint32_t            subCode;          /**< device product sub code */
  uint32_t            revision;         /**< any specific firmware revision, 0 = generic/all */
  uint32_t            serialNumber;     /**< device serial number */

  ec_slave_config_t  *slaveConfig;      /**< slave configuration */
  
  // methods
  ecDevice_fn         prepareToRun;     /**< prepare the device to run */
  ecDevice_fn         run;              /**< run device logic */
  ecDevice_fn         prepareToStop;    /**< prepare the device to stop, returns 1 when ready to stop */
};



/** create an ethercat device
 *
 * \retval 0 on success else error code
 */
int create_etherCATDevice(
        etherCATMaster_s *in_parent,            /**< the parent master */
        ecDeviceInfo_t *in_deviceInfo,          /**< the device information */
        uint32_t in_revision,                   /**< an optional specific revision of firmware */
        uint16_t in_alias,                      /**< device alias */
        uint16_t in_position,                   /**< device position */
        ecDevice_s **out_device                 /**< a pointer to the new device */
        );


typedef struct _ecAxis_s
{
  ecDevice_s *m_device;
} ecAxis_s;


/** Yaskawa SGDV axis device
 */
typedef struct _yaskawaSGDVAxis_s
{
  ecAxis_s m_axis;

  // commands
  ecItem_s cmdControl;                                                          // U16
  ecItem_s   cmdControl_enableSwitchOn;
  ecItem_s   cmdControl_enableVoltage;
  ecItem_s   cmdControl_enableNotQuickStop;
  ecItem_s   cmdControl_enableOperation;
  ecItem_s   cmdControl_faultReset;
  ecItem_s cmdTargetPos;
  ecItem_s cmdTargetVel;
  ecItem_s cmdMaxTorqueLim;
  ecItem_s cmdOutputs;
  ecItem_s cmdLatchControl;
  ecItem_s cmdSelectMode;                                                       // S8

  ecItem_s cmdStatus;
  ecItem_s   cmdStatus_enableReadyToSwitchOn;
  ecItem_s   cmdStatus_enableSwitchedOn;
  ecItem_s   cmdStatus_enableOperationEnabled;
  ecItem_s   cmdStatus_fault;
  ecItem_s cmdActualPos;
  ecItem_s cmdActualVel;
  ecItem_s cmdActualTorque;
  ecItem_s cmdPosError;
  ecItem_s cmdInputs;
  ecItem_s cmdLatchPos;
  ecItem_s cmdLatchStatus;
  ecItem_s cmdCurrentMode;
  ecItem_s cmdHomingMethod;
  ecItem_s cmdLastError;
  
} yaskawaSGDVAxis_s;



/** Yaskawa SGDV device
 */
typedef struct _yaskawaSGDVDevice_s
{
  ecDevice_s  m_device;
  
  yaskawaSGDVAxis_s  axis0;           /**< axis channel */
  
} yaskawaSGDVDevice_s;



/** domain groups */
#define DOM_MAIN       0
#define DOM_AXIS_READ  1
#define DOM_AXIS_WRITE 2


typedef enum
{
  maError,          /**< an error occured activating the master */
  maInactive,       /**< master is not active */
  maActive,         /**< master is now active */
  maDeactivating    /**< master deactivate deactivate slaves has been called */
} masterActivateState_e;


#define EC_DOMAIN_COUNT 3


/** master function
 *
 * \retval 0 on success, else error code
 */
typedef int (*ecMaster_fn)(
        void *this               /**< the master struct */
        );



/** The EtherCAT master data structure
 */
struct _etherCATMaster_s
{
  ec_master_t            *master;               /**< the ethercat master reference once attached */
  ec_master_state_t       masterState;          /**< cached master state */
  int32_t                 checkMasterCntr;      /**< check master state counter */
  masterActivateState_e   masterActivateState;  /**< master activate control state */
          
  etherCATDomain_t        domains[EC_DOMAIN_COUNT]; /**< the domains */
  
  ecDevice_s            **devices;              /**< the configured EtherCAT devices */
  int                     deviceCount;          /**< the number of devices */

  ecDevice_s             *m_dcRefSlave;         /**< is there a configured ref slave? */
  uint64_t                m_dcTimeStart;        /**< distributed clock first now time */
  uint64_t                m_dcTime;             /**< distributed clock now time */
  int32_t                 m_dcDiff;             /**< distributed clock master to ref slave diff time */
  bool                    m_getDCDiff;          /**< we are ready to get the distributed clock diff time */
  int                     m_lastDCRes;          /**< track the last return value of ecrt_master_reference_clock_time */

  ecMaster_fn             prepareToRun;         /**< prepare the master to run */
  ecMaster_fn             run;                  /**< process master */
  ecMaster_fn             prepareToStop;        /**< prepare the master to stop */
  ecMaster_fn             stop;                 /**< stop the master */
};



/** configured devices
 */
typedef struct _ecDeviceList_t
{
  char         *deviceTypeName;       /**< device type name */
  uint16_t      alias;                /**< device alias */
  uint16_t      position;             /**< device position */
  bool          dcRefSlave;           /**< use this slave as the dc clock ref slave */
} ecDeviceList_t;



/** Calc the EtherCAT item data address
 */
void setECAddress(
        ecItem_s *ecItem
        );



/** create / initialise an EtherCAT module
 *
 * \retval 0 on success else error code
 */
int create_etherCATMaster(
        ecDeviceList_t *in_devices,     /**< list of devices to configure */
        void **out_master               /**< a pointer to the new master */
        );



#endif	/* _ETHERCAT_MASTER_H */

