/*
 * etherCATSlaves.c
 *
 * implementation of EtherCAT slaves
 *
 * 23/08/2017    GF    Created
 * 
 * 
 * MIT License
 * 
 * Copyright 2017 Graeme Foot (Kinetic Engineering Design Ltd.)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <stddef.h>
#include <ecrt.h>
#include <limits.h>

#include "data.h"
#include "etherCATMaster.h"


/** get the slave information and absolute ring position
 */
int ecDevice_getSlaveInfo(
        void *this,                         /**< the ec device ecDevice_s */
        uint16_t *out_ringPosition,         /**< absolute ring position */
        ec_slave_info_t *out_slaveInfo,     /**< slave information, also used as scratchpad */
        bool *out_found                     /**< whether a match was found */
        )
{
  ecDevice_s       *dev    = this;
  ec_master_t      *master = dev->m_parent->master;
  ec_master_info_t  master_info;
  int               res;
  int               i;
  uint16_t          offset = 0;

  // we don't store the absolute ring position because of hot swapping
  // we need to look it up by matching against the alias/position pair
  res = ecrt_master(master, &master_info);
  if (res != 0)
  {
    logMsg("getSlaveInfo: Error getting master info");
    return -1;
  }
  
  
  if (dev->alias == 0)
  {
    // lookup by absolute position
    if (dev->position < master_info.slave_count)
    {
      res = ecrt_master_get_slave(master, dev->position, out_slaveInfo);
      if (res != 0)
      {
        logMsg("getSlaveInfo: Error getting slave info");
        return -1;
      }

      *out_ringPosition = dev->position;
      *out_found        = true;

      return 0;
    }
  }
  else
  {
    // lookup by alias / offset position
    for (i = 0; i < master_info.slave_count; i++)
    {
      res = ecrt_master_get_slave(master, i, out_slaveInfo);
      if (res != 0)
      {
        logMsg("getSlaveInfo: Error getting slave info");
        return -1;
      }

      if (dev->alias == out_slaveInfo->alias)
      {
        if (dev->position == offset)
        {
          *out_ringPosition = i;
          *out_found        = true;

          return 0;
        }
        else
        {
          offset++;
        }
      }
    }
  }

  *out_found = false;
  
  return 0;
}



/** fill in slave info for the device (if its available)
 * 
 * and check that we are about to configure the correct slave
 * ie: vendorID, productCode and optionally revision matches
 */
int ecDevice_fillInSlaveInfo(
        void *this                          /**< the ec device ecDevice_s */
        )
{
  ecDevice_s      *dev = this;
  uint16_t         ringPosition;
  ec_slave_info_t  slaveInfo;
  bool             found;

  if (ecDevice_getSlaveInfo(this, &ringPosition, &slaveInfo, &found) != 0) return -1;
  
  if (found)
  {
    if ( (dev->vendorID != slaveInfo.vendor_id) ||
         (dev->productCode != slaveInfo.product_code) )
    {
      logMsg("Slave Vendor ID or Product Code mismatch");
      return -1;
    }
    if ( (dev->revision != 0) && 
         (dev->revision != slaveInfo.revision_number) ) 
    {
      logMsg("Slave revision mismatch");
      return -1;
    }
    
    // fill in revision and serial number (not that is seems to be filled in)
    dev->revision     = slaveInfo.revision_number;
    dev->serialNumber = slaveInfo.serial_number;
  }
  
  
  return 0;
}




// helpers /////////////////////////////////////////////////////////////////////

/** helper to create ec device, setting up its pdo's and preparing it for create_ecCmd calls
 * it requires the following to be pre defined:
 *   ecDevice_s       *dev
 *   etherCATModule_s *ecMaster
 */
int ecDevice_init(
        etherCATMaster_s *in_ecMaster, 
        ecDevice_s *in_ecDev, 
        ec_sync_info_t in_pdoSyncs[]
        )
{
  int ecRes;
  
  // configure the slave 
  in_ecDev->slaveConfig = ecrt_master_slave_config(in_ecMaster->master, in_ecDev->alias, in_ecDev->position,
                                                   in_ecDev->vendorID, in_ecDev->productCode);
  if (in_ecDev->slaveConfig == NULL)
  {
    logMsg("Error getting slave configuration");
    return -1;
  }
  
  // configure the PDOs 
  if (in_pdoSyncs != NULL)
  {
    ecRes = ecrt_slave_config_pdos(in_ecDev->slaveConfig, EC_END, in_pdoSyncs);
    if (ecRes != 0)
    {
      logMsg("Error configuring slave pdos");
      return -1;
    }
  }

  return 0;
}



/** helper to create pdo entry ec commands
 */
#define create_ecItem(_ecMaster, _ecDev, _pdoEntries, _pdoIdx, _name, _domIdx)                        \
({                                                                                                    \
  int           offset;                                                                               \
  unsigned int  bitPos;                                                                               \
  ecItem_s      ecItem;                                                                               \
                                                                                                      \
  offset = ecrt_slave_config_reg_pdo_entry((_ecDev)->slaveConfig, (_pdoEntries)[_pdoIdx].index,       \
                                           (_pdoEntries)[_pdoIdx].subindex,                           \
                                           (_ecMaster)->domains[_domIdx].domain, &bitPos);            \
                                                                                                      \
  if (offset < 0)                                                                                     \
  {                                                                                                   \
    logMsg("Error getting PDO entry");                                                                \
    return -1;                                                                                        \
  }                                                                                                   \
                                                                                                      \
  ecItem.name   = (_name);                                                                            \
  ecItem.master = (_ecMaster);                                                                        \
  ecItem.domain = (_domIdx);                                                                          \
  ecItem.offset = offset;                                                                             \
  ecItem.bitPos = bitPos;                                                                             \
  ecItem.addr   = NULL;                                                                               \
                                                                                                      \
  ecItem;                                                                                             \
})



/** helper to create pdo entry to a bit in an existing command
 */
#define set_ecItemBitPos(_ecItem, _bitPos, _name)                             \
({                                                                            \
  ecItem_s      ecItem;                                                       \
                                                                              \
  ecItem.name   = (_name);                                                    \
  ecItem.master = (_ecItem).master;                                           \
  ecItem.domain = (_ecItem).domain;                                           \
  ecItem.offset = (_ecItem).offset;                                           \
  ecItem.bitPos = _bitPos;                                                    \
  ecItem.addr   = NULL;                                                       \
                                                                              \
  ecItem;                                                                     \
})



/** blank device function
 *
 * \retval 0 on success, else error code
 */
static int device_void(
        void *this                      /**< the device ecDevice_s */
        )
{
  return 0;
}


// EK1110 //////////////////////////////////////////////////////////////////////

/** create the EtherCAT device
 */
static int EK1110_create(
        void *this                      /**< the device ecDevice_s */
        )
{
  ecDevice_s       *dev      = this;
  etherCATMaster_s *ecMaster = dev->m_parent;
  

  // set device methods
  dev->prepareToRun    = device_void;
  dev->run             = device_void;
  dev->prepareToStop   = device_void;
  
  // fill in information for the slave
  if (ecDevice_fillInSlaveInfo(dev) != 0) return -1;
  
  
  // init pdo's and prepare for cmd objects
  if (ecDevice_init(ecMaster, dev, NULL) != 0) return -1;
  
  
  return 0;
}



// yaskawaSGDV ////////////////////////////////////////////////////////////////

/** max counts per cycle for target position
 *  actual yaskawa limit value = 4194304
 *  Note: divided by 3 so that if a frame is missed it won't go over the limit
 *        on the next frame 
 *        (by 3 because the amp will error out via 0x10F1:02 first) */
#define MAX_YASK_CYCLE_DELTA 4194000 / 3


// Object dictionary on page 8-2 of EtherCAT users manual

ec_pdo_entry_info_t yaskawaSGDV_pdoEntries[] = {
    {0x6040, 0x00, 16},   // pg 8-21  control (rw)
    {0x607a, 0x00, 32},   // pg 8-30  target position (rw) (for cyclic sync pos mode)
    {0x60b1, 0x00, 32},   // pg 8-37  velocity offset (rw) (for cyclic sync pos mode)
    {0x6072, 0x00, 16},   // pg 8-40  max torque (rw) in 0.1% increments
    {0x60fe, 0x01, 32},   // pg 8-44  outputs (rw) (bits 17, 18, 19 control outputs 1, 2, 3)
    {0x60b8, 0x00, 16},   // pg 8-41  latchControl (rw) (bits: 0 enable latch, 1 single/continuous trigger,
                          //                             2 0=latch on S14 | 1=latch on C, 4 enable latch sampling???)
    {0x6060, 0x00, 8},    // pg 8-28  mode of operation (rw) (8=cyclic pos, 9=cyclic vel)

    {0x6041, 0x00, 16},   // pg 8-23  status (r)
    {0x6064, 0x00, 32},   // pg 8-34  actual position (r)
    {0x606c, 0x00, 32},   // pg 8-38  actual velocity (r)
    {0x6077, 0x00, 16},   // pg 8-39  actual torque (r)
    {0x60f4, 0x00, 32},   // pg 8-35  actual following error (r)
    {0x60fd, 0x00, 32},   // pg 8-43  inputs (r) (available inputs: 0 (rev lim), 1 (fwd lim), 2 (hm), 
                          //                      16 - 22 (via CN1), 25 - 25 (via base block))
    {0x60ba, 0x00, 32},   // pg 8-42  latchPos (r) (via touch probe 1)
    {0x60b9, 0x00, 16},   // pg 8-42  latchStatus (r) (bits: 0 latch enabled, 1 latched, 7 latched toggle (for continuous mode))
    {0x6061, 0x00, 8},    // pg 8-28  current mode of operation
    {0x6098, 0x00, 8},    // pg 8-32  homing mode (rw) note: rw but we only want to read current value (setting new value by sdo)
    {0x603f, 0x00, 16},   // pg 8-21  last error code
    {0x6062, 0x00, 32},   // pg 8-34  position demand value (r)
};

ec_pdo_info_t yaskawaSGDV_pdos[] = {
    {0x1600, 7, yaskawaSGDV_pdoEntries + 0},
    {0x1a00, 8, yaskawaSGDV_pdoEntries + 7},
    {0x1a01, 4, yaskawaSGDV_pdoEntries + 15},
};

ec_sync_info_t yaskawaSGDV_syncs[] = {
    {0, EC_DIR_OUTPUT, 0, NULL, EC_WD_DISABLE},
    {1, EC_DIR_INPUT, 0, NULL, EC_WD_DISABLE},
    {2, EC_DIR_OUTPUT, 1, yaskawaSGDV_pdos + 0, EC_WD_ENABLE},
    {3, EC_DIR_INPUT, 2, yaskawaSGDV_pdos + 1, EC_WD_DISABLE},
    {0xff}
};



/** prepare device to run
 *
 * \retval 0 on success, else error code
 */
static int yaskawaSGDV_prepareToRun(
        void *this                      /**< the device ecDevice_s */
        )
{
  yaskawaSGDVDevice_s *yask     = this;
  yaskawaSGDVAxis_s   *yaskAxis = &yask->axis0;
  
  
  // calculate cmd addresses once
  setECAddress(&yaskAxis->cmdControl); 
  setECAddress(&yaskAxis->cmdControl_enableSwitchOn);
  setECAddress(&yaskAxis->cmdControl_enableVoltage);
  setECAddress(&yaskAxis->cmdControl_enableNotQuickStop);
  setECAddress(&yaskAxis->cmdControl_enableOperation);
  setECAddress(&yaskAxis->cmdControl_faultReset);
  setECAddress(&yaskAxis->cmdTargetPos);
  setECAddress(&yaskAxis->cmdTargetVel);
  setECAddress(&yaskAxis->cmdMaxTorqueLim);
  setECAddress(&yaskAxis->cmdOutputs);
  setECAddress(&yaskAxis->cmdLatchControl);
  setECAddress(&yaskAxis->cmdSelectMode);  

  setECAddress(&yaskAxis->cmdStatus);
  setECAddress(&yaskAxis->cmdStatus_enableReadyToSwitchOn);
  setECAddress(&yaskAxis->cmdStatus_enableSwitchedOn);
  setECAddress(&yaskAxis->cmdStatus_enableOperationEnabled);
  setECAddress(&yaskAxis->cmdStatus_fault);
  setECAddress(&yaskAxis->cmdActualPos);
  setECAddress(&yaskAxis->cmdActualVel);
  setECAddress(&yaskAxis->cmdActualTorque);
  setECAddress(&yaskAxis->cmdPosError);
  setECAddress(&yaskAxis->cmdInputs);
  setECAddress(&yaskAxis->cmdLatchPos);
  setECAddress(&yaskAxis->cmdLatchStatus);
  setECAddress(&yaskAxis->cmdCurrentMode);
  setECAddress(&yaskAxis->cmdHomingMethod);
  setECAddress(&yaskAxis->cmdLastError);
  
  
  
  // set mode to Cyclic Synchronous Position  8
  EC_WRITE_S8(yaskAxis->cmdSelectMode.addr, 8);

  
  // reset the control word
  EC_WRITE_U16(yaskAxis->cmdControl.addr, 0);

  
  return 0;
}



/** run device
 *
 * \retval 0 on success, else error code
 */
static int yaskawaSGDV_run(
        void *this                      /**< the device ecDevice_s */
        )
{
  // *** Add cyclical run code here ***
  
  return 0;
}



/** prepare to stop device
 *
 * \retval 0 on success, else error code
 */
static int yaskawaSGDV_prepareToStop(
        void *this                      /**< the device ecDevice_s */
        )
{
  return 0;
}



/** create the EtherCAT device (pre-allocated memory, alias and position)
 */
static int yaskawaSGDV_create(
        void *this                 /**< the device ecDevice_s */
        )
{
  ecDevice_s          *dev      = this;
  yaskawaSGDVDevice_s *yask     = this;
  yaskawaSGDVAxis_s   *yaskAxis = &yask->axis0;
  ecAxis_s            *ecAxis   = (ecAxis_s *)yaskAxis;

  etherCATMaster_s    *ecMaster = dev->m_parent;
  
  // config the axis channel
  ecAxis->m_device     = dev;
  
  // set device methods
  dev->prepareToRun    = yaskawaSGDV_prepareToRun;
  dev->run             = yaskawaSGDV_run;
  dev->prepareToStop   = yaskawaSGDV_prepareToStop;
  
  // fill in information for the slave
  if (ecDevice_fillInSlaveInfo(dev) != 0) return -1;
  
  
  // init pdo's and prepare for cmd objects
  if (ecDevice_init(ecMaster, dev, yaskawaSGDV_syncs) != 0) return -1;
  

  // get pdo entries
  yaskAxis->cmdControl      = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 0, "control", DOM_AXIS_WRITE);
  yaskAxis->cmdTargetPos    = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 1, "targetPos", DOM_AXIS_WRITE);
  yaskAxis->cmdTargetVel    = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 2, "targetVel", DOM_AXIS_WRITE);
  yaskAxis->cmdMaxTorqueLim = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 3, "posTorqueLim", DOM_AXIS_WRITE);
  yaskAxis->cmdOutputs      = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 4, "outputs", DOM_AXIS_WRITE);
  yaskAxis->cmdLatchControl = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 5, "latchControl", DOM_AXIS_WRITE);
  yaskAxis->cmdSelectMode   = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 6, "selectMode", DOM_AXIS_WRITE);

  yaskAxis->cmdStatus       = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 7, "status", DOM_AXIS_READ);
  yaskAxis->cmdActualPos    = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 8, "actualPos", DOM_AXIS_READ);
  yaskAxis->cmdActualVel    = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 9, "actualVel", DOM_AXIS_READ);
  yaskAxis->cmdActualTorque = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 10, "actualTorque", DOM_AXIS_READ);
  yaskAxis->cmdPosError     = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 11, "posError", DOM_AXIS_READ);
  yaskAxis->cmdInputs       = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 12, "inputs", DOM_AXIS_READ);
  yaskAxis->cmdLatchPos     = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 13, "latchPos", DOM_AXIS_READ);
  yaskAxis->cmdLatchStatus  = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 14, "latchStatus", DOM_AXIS_READ);
  yaskAxis->cmdCurrentMode  = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 15, "currentMode", DOM_AXIS_READ);
  yaskAxis->cmdHomingMethod = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 16, "homingMethod", DOM_AXIS_READ);
  yaskAxis->cmdLastError    = create_ecItem(ecMaster, dev, yaskawaSGDV_pdoEntries, 17, "lastError", DOM_AXIS_READ);

  
  // create the control bits
  yaskAxis->cmdControl_enableSwitchOn     = set_ecItemBitPos(yaskAxis->cmdControl, 0, "enable_switchOn");
  yaskAxis->cmdControl_enableVoltage      = set_ecItemBitPos(yaskAxis->cmdControl, 1, "enable_voltage");
  yaskAxis->cmdControl_enableNotQuickStop = set_ecItemBitPos(yaskAxis->cmdControl, 2, "enable_notQuickStop");
  yaskAxis->cmdControl_enableOperation    = set_ecItemBitPos(yaskAxis->cmdControl, 3, "enable_operation");
  yaskAxis->cmdControl_faultReset         = set_ecItemBitPos(yaskAxis->cmdControl, 7, "faultReset");

  // create the status bits
  yaskAxis->cmdStatus_enableReadyToSwitchOn  = set_ecItemBitPos(yaskAxis->cmdStatus, 0, "enable_readyToSwitchOn");
  yaskAxis->cmdStatus_enableSwitchedOn       = set_ecItemBitPos(yaskAxis->cmdStatus, 1, "enable_switchedOn");
  yaskAxis->cmdStatus_enableOperationEnabled = set_ecItemBitPos(yaskAxis->cmdStatus, 2, "enable_operationEnabled");
  yaskAxis->cmdStatus_fault                  = set_ecItemBitPos(yaskAxis->cmdStatus, 3, "fault");
  
  
  return 0;
}



/** a library of known devices and any device specific functions.
 * if a function is null then it will use the default ecDevice_... version
 * 
 * Note: the last entries name in the array should be NULL 
 *       (so we don't need to know the count)
 */
ecDeviceInfo_t ecDeviceTypes[] =
{
  // beckhoff
  {"EK1110",       0x00000002, 0x04562c52, 0, sizeof(ecDevice_s), EK1110_create},
  
  // yaskawa
  {"Yaskawa-SGDV", 0x00000539, 0x02200001, 0, sizeof(yaskawaSGDVDevice_s), yaskawaSGDV_create},
  
  {NULL}
};


