/*
 * main.c
 *
 * main application entry point
 *
 * 23/08/2017    GF    Created
 * 
 * 
 * MIT License
 * 
 * Copyright 2017 Graeme Foot (Kinetic Engineering Design Ltd.)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */



#include <signal.h>
#include <execinfo.h>


#include "data.h"
#include "etherCATMaster.h"



/**
 * the global application data
 */
app_s g_app = {.shutdown    = 0,
               .scanTimeNS  = 1000000ull,
               .scanTimeMS  = 1ull,
               .scanTimeSec = 0.001,
               .rtTimeDiff  = 0,
               .rtTimeDelta = 0,
               .rtTimeStart = 0LL,
               .rtTimeCurr  = 0LL,
               .rtAdjBy     = 0LL};
               
#define APPTSK "APPTSK"


/**
 * set the exit app flag
 */
void endApp(
        int sig         /**< exit signal */
        )
{
  // ignore
  if (sig == SIGCHLD) 
  {
    return;
  }
  
  
  if (sig == 0)
  {
    g_app.shutdown = 1;
    return;
  }
  
  
  // ignore:
  // - SIGHUP
  // - SIGPIPE
  
  if ((sig == SIGTERM) || (sig == SIGINT) || 
      (sig == SIGKILL) || (sig == SIGALRM))
  {
    if (((sig == SIGINT) || (sig == SIGKILL)) && g_app.shutdown)
    {
      logMsg("HALTED\n");
      exit(128 + sig);
    }

    if (sig == SIGTERM) logMsg("Shutting down\n");
    else                logMsg("Shutting down\n");
    
    g_app.shutdown = true;
  }
  
  // SIGSEGV && SIGBUS are access violations
  // SIGABRT is called by assert
  else if ((sig == SIGSEGV) || (sig == SIGBUS) || (sig == SIGABRT))
  {
    logMsg("APP HALT ERROR %d (%s), backtrace:\n", sig, strsignal(sig));
    
    // output backtrace
    void   *buffer[100];
    int     cnt;
    int     i;
    
    // get void*'s for all entries on the stack
    cnt = backtrace(buffer, 100);
    
    // print out frames to stderr
    backtrace_symbols_fd(buffer, cnt, STDERR_FILENO);
    
    // print out frames to syslog
    char **strings = backtrace_symbols(buffer, cnt);
    logMsg(" %d frames:\n", cnt);
    for (i = 0; i < cnt; i++)
    {
      logMsg("  %s\n", strings[i]);
    }
    logMsg("backtrace done.\n");
    
    free(strings);
    
    
    g_app.shutdown = true;

    // exit
    exit(128 + sig);
  }
}



/**
 * ECat thread entry point 
 */
int runECat(void *args)
{
  // configure devices
  ecDeviceList_t ecatDevices[] =
  {
    // device          alias  position  dc ref slave
    {  "EK1110",       10001, 0,        false},
    {  "Yaskawa-SGDV", 12,    0,        true},
    {NULL}
  };

  
  // initialise EtherCAT master
  if (create_etherCATMaster(ecatDevices, (void **)&g_app.ecatMaster) != 0) 
  {
    exit(EXIT_ERROR);
  }
  
  if (g_app.ecatMaster->prepareToRun(g_app.ecatMaster) != 0)
  {
    logMsg("Error during prepareToRun");
    exit(EXIT_ERROR);
  }


  // initialise the thread task
  if (!(g_app.appTask = rt_thread_init(nam2num(APPTSK), 0/*priority*/, 0, SCHED_FIFO, 0x2)))
  {
    logMsg("Error initialising rt thread");
    exit(EXIT_ERROR);
  }

  // set timer mode
  rt_set_oneshot_mode();
  
  // allow non root user to lock memory
  rt_allow_nonroot_hrt();
  mlockall(MCL_CURRENT | MCL_FUTURE);

  // start at nearest round period 50 periods in the future
  uint64_t firstWakeTime = app_getTimeNS();
  firstWakeTime -= (firstWakeTime % g_app.scanTimeNS);
  firstWakeTime += 50*g_app.scanTimeNS;
  
  // start the timer
  start_rt_timer(nano2count(g_app.scanTimeNS));
  g_app.nextWakeTime   = firstWakeTime;
  g_app.actualWakeTime = 0;
  
  rt_make_hard_real_time();

  
  // initial hard sleep
  ecat_waitPeriod();
  
  // main loop
  while (!g_app.shutdown)
  {
    // run master
    if (g_app.ecatMaster->run(g_app.ecatMaster) != 0)
    {
      logMsg("Error during run");
      exit(EXIT_ERROR);
    }

    // polling hard sleep
    ecat_waitPeriod();
  }
  
  
  // shutdown
  if (g_app.ecatMaster->prepareToStop(g_app.ecatMaster) != 0)
  {
    logMsg("Error during prepareToStop");
    exit(EXIT_ERROR);
  }
  
  
  // all done, go soft again
  rt_make_soft_real_time();
  stop_rt_timer();
  rt_task_delete(g_app.appTask);
  
  
  // cleanup
  if (g_app.ecatMaster->stop(g_app.ecatMaster) != 0)
  {
    logMsg("Error during stop");
    exit(EXIT_ERROR);
  }
  
  
  return 0;
}


/**
 * Main server application entry point
 */
int main(int argc, char *argv[])
{
  // defines
#ifdef DEBUG
  bool defined_DEBUG;
  bool debug_build;
#endif


  // set up signal handlers
  // for debugging, set all
  int i;
  for (i = 1; i < _NSIG; i++)
  {
    signal(i, endApp);
  }
  

  // start the EtherCAT realtime thread
  long ecatThreadID = rt_thread_create(runECat, NULL, 10000);


  // wait for shutdown flag to be set (once every 5 seconds)
  while (!g_app.shutdown)
  {
    sleep(5);
  }


  rt_thread_join(ecatThreadID);

  
  return 0;
}


